import pymunk

from .settings import NODE_RADIUS
from .visual_graph import VisualGraph

VERTEX_BODY_MASS = 1
VERTEX_BODY_MOMENT = 1
WALLS_WIDTH = 10


def create_border(space, bounds: (int, int)):
    width, height = bounds
    # Create the ground (static) segment
    ground = pymunk.Segment(space.static_body, (0, -WALLS_WIDTH), (width, -WALLS_WIDTH), WALLS_WIDTH)
    ground.friction = 1.0
    space.add(ground)

    # Create the ceiling (static) segment
    ceiling = pymunk.Segment(space.static_body, (0, height + WALLS_WIDTH), (width, height + WALLS_WIDTH), WALLS_WIDTH)
    ceiling.friction = 1.0
    space.add(ceiling)

    # Create the left (static) segment
    left_wall = pymunk.Segment(space.static_body, (-WALLS_WIDTH, 0), (-WALLS_WIDTH, height), WALLS_WIDTH)
    left_wall.friction = 1.0
    space.add(left_wall)

    # Create the right (static) segment
    right_wall = pymunk.Segment(space.static_body, (width + WALLS_WIDTH, 0), (width + WALLS_WIDTH, height), WALLS_WIDTH)
    right_wall.friction = 1.0
    space.add(right_wall)


class GraphPhysics:
    def __init__(self, visual_graph: VisualGraph):
        self.visual_graph = visual_graph
        self.space = pymunk.Space()
        create_border(self.space, visual_graph.bounds)
        self.space.gravity = (0, 0)
        self.vertex_body: dict[any, pymunk.Body] = {}
        self.edge_body: dict[any, pymunk.Body] = {}
        self.connection_body = {}
        for node in visual_graph.graph.nodes:
            self.add_vert(node, visual_graph.coordinates[node])
        for edge in visual_graph.graph.edges:
            self.add_edge(edge[0], edge[1])
        visual_graph.add_node.subscribable.subscribe(self.add_vert)
        visual_graph.remove_node.subscribable.subscribe(self.remove_vert)
        visual_graph.add_edge.subscribable.subscribe(self.add_edge)
        visual_graph.remove_edge.subscribable.subscribe(self.remove_edge)
        visual_graph.move_node.subscribable.subscribe(self.move_node)

        visual_graph.drag_start.subscribable.subscribe(self.drag_start)
        visual_graph.drag_end.subscribable.subscribe(self.drag_end)

    def drag_start(self, node):
        self.vertex_body[node].body_type = pymunk.Body.STATIC

    def drag_end(self):
        node = self.visual_graph.dragged_node
        if node is not None:
            self.vertex_body[node].body_type = pymunk.Body.DYNAMIC
            self.vertex_body[node].mass = VERTEX_BODY_MASS
            self.vertex_body[node].moment = VERTEX_BODY_MOMENT

    def add_vert(self, node, pos: (int, int)):
        body = pymunk.Body(VERTEX_BODY_MASS, VERTEX_BODY_MOMENT)
        body.position = pos
        shape = pymunk.Circle(body, radius=10)  # Adjust the radius as needed
        shape.elasticity = 1.0  # Elasticity of collisions
        shape.friction = 0.0  # Friction of collisions
        self.space.add(body, shape)
        self.vertex_body[node] = body
        self.edge_body[node] = {}
        self.connection_body[node] = {}
        for node1, body1 in self.vertex_body.items():
            if node1 != node:
                connection_body = pymunk.DampedSpring(body, body1, (0, 0), (0, 0), rest_length=200, stiffness=10,
                                                      damping=2)
                self.space.add(connection_body)
                self.connection_body[node][node1] = connection_body
                self.connection_body[node1][node] = connection_body

    def remove_vert(self, node):
        self.space.remove(self.vertex_body[node])
        del self.vertex_body[node]
        for other in self.edge_body[node]:
            self.space.remove(self.edge_body[node][other])
            del self.edge_body[other][node]
        for other in self.connection_body[node]:
            self.space.remove(self.connection_body[node][other])
            del self.connection_body[other][node]
        del self.edge_body[node]
        del self.connection_body[node]

    def add_edge(self, node1, node2):
        edge_body = pymunk.DampedSpring(self.vertex_body[node1], self.vertex_body[node2], (0, 0), (0, 0),
                                        rest_length=100, stiffness=500, damping=2)
        self.space.add(edge_body)
        self.edge_body[node1][node2] = edge_body
        self.edge_body[node2][node1] = edge_body

    def remove_edge(self, node1, node2):
        body = self.edge_body[node1][node2]
        self.space.remove(body)
        del self.edge_body[node1][node2]
        del self.edge_body[node2][node1]

    def move_node(self, node, pos: (int, int)):
        self.vertex_body[node].position = pos

    def update_physics(self, dt, physics):
        if physics:
            self.space.step(dt)
            for node, body in self.vertex_body.items():
                self.visual_graph.move_node(node, [body.position.x, body.position.y])
        self.normalize_positions()

    def normalize_positions(self):
        for node, node_pos in self.visual_graph.coordinates.items():
            if node_pos[0] < 0:
                self.visual_graph.move_node(node, [NODE_RADIUS + 10, node_pos[1]])
                self.vertex_body[node].velocity = [0, self.vertex_body[node].velocity.y]
            if node_pos[1] < 0:
                self.visual_graph.move_node(node, [node_pos[0], NODE_RADIUS + 10])
                self.vertex_body[node].velocity = [self.vertex_body[node].velocity.x, 0]
            if node_pos[0] > self.visual_graph.bounds[0]:
                self.visual_graph.move_node(node, [self.visual_graph.bounds[0] - NODE_RADIUS - 10, node_pos[1]])
                self.vertex_body[node].velocity = [0, self.vertex_body[node].velocity.y]
            if node_pos[1] > self.visual_graph.bounds[1]:
                self.visual_graph.move_node(node, [node_pos[0], self.visual_graph.bounds[1] - NODE_RADIUS - 10])
                self.vertex_body[node].velocity = [self.vertex_body[node].velocity.x, 0]



from ipycanvas import Canvas, hold_canvas
from .settings import DRAGGED_NODE_RADIUS, NODE_RADIUS
from .visual_graph import VisualGraph
import ipywidgets as widgets


def draw_graph(canvas: Canvas, visual_graph: VisualGraph):
    def clear_canvas():
        canvas.clear()
        canvas.stroke_style = "black"
        canvas.stroke_rect(0, 0, 800, 500)

    def draw_vertex(pos, size=10, colorcode="black"):
        canvas.fill_style = colorcode
        canvas.fill_circle(pos[0], pos[1], size)

    def draw_edge(pos1, pos2, colorcode="black"):
        canvas.stroke_style = colorcode
        canvas.line_width = 2
        canvas.stroke_line(*pos1, *pos2)

    with hold_canvas():
        clear_canvas()
        for edge in visual_graph.graph.edges:
            draw_edge(visual_graph.coordinates[edge[0]], visual_graph.coordinates[edge[1]],
                      colorcode=("red" if edge == visual_graph.selected_edge else "black"))
        for node, pos in visual_graph.coordinates.items():
            draw_vertex(pos,
                        size=(DRAGGED_NODE_RADIUS if node == visual_graph.dragged_node else NODE_RADIUS),
                        colorcode=("red" if node == visual_graph.selected_node else "black"))


class SmallButton(widgets.Button):
    def __init__(self, active=False, active_color=None, inactive_color=None, description='',**kwargs):
        super().__init__(layout=widgets.Layout(width='39px', height='39px'), description=description, **kwargs)
        self.active=active
        self.active_color=active_color
        self.inactive_color=inactive_color
        self.style.button_color = active_color if active else inactive_color

    def toggle(self):
        self.active=not self.active
        self.style.button_color = self.active_color if self.active else self.inactive_color


class Menu(widgets.HBox):
    def __init__(self):
        super().__init__()
        self.close_button = SmallButton(tooltip='Exit', icon='window-close')
        self.physics_button = widgets.ToggleButton(
            value=True,
            tooltip='Turn physics on/off',
            disabled=False,
            indent=False,
            layout=widgets.Layout(width='39px', height='39px'), icon="wrench")

        self.struct_button = SmallButton(tooltip='Click to activate edges and vertices creation/deletion',
                                            icon='plus-circle', active_color='LightBlue', active=True)


        self.prop_button = SmallButton(tooltip='Click to modify properties of edges and vertices',
                                       active_color='LightBlue', icon="pencil")

        self.edge_button = SmallButton(tooltip='Edges selection enabled/disabled', 
                                          icon='arrows-v', active_color='LightGreen', inactive_color='lightcoral', active=True)

        self.vert_button = SmallButton(tooltip='Vertices selection enabled/disabled',
                                        icon='circle', active_color='LightGreen', inactive_color='lightcoral', active=True)

        self.children = ([widgets.HBox((self.struct_button, self.prop_button),
                                       layout=widgets.Layout(border='0.5px solid #000000')),
                          self.vert_button, self.edge_button, self.physics_button, self.close_button])


def get_label_style():
    return dict(
        font_weight='bold',
        background='#d3d3d3',
        font_variant="small")


class LabelBox(widgets.HBox):
    def __init__(self, label_value, text_value):
        super().__init__()
        self.label_value = widgets.Textarea(value=text_value,
                                            layout=widgets.Layout(width='100px', height='30px'))

        label_label = widgets.Label(value=label_value,
                                    layout=widgets.Layout(width='150px', height='30px'), style=get_label_style())

        label_label.layout.border = '2px solid #000000'
        self.children = (label_label, self.label_value)



class LabelListBox(widgets.VBox):
    text_layout = widgets.Layout(width='180px', height='35px')
    button_layout = widgets.Layout(width='35px', height='35px')
    def __init__(self, str_value):
        super().__init__()
        self.label = widgets.Label(value=str_value,
                                   layout=self.text_layout,
                                   style=get_label_style())
        self.label.layout.border = '2px solid #000000'
        self.delete_button = widgets.Button(layout=self.button_layout, icon="trash-o")
        self.edit_button = widgets.Button(layout=self.button_layout, icon="pencil")
        
        self.edit_label_value=widgets.Textarea(placeholder='New label name', layout=self.text_layout)
        self.confirm_edit_button=widgets.Button(layout=self.button_layout, icon="check")
        self.escape_edit_button=widgets.Button(layout=self.button_layout, icon='times')

        self.children=(widgets.HBox((self.label, self.edit_button, self.delete_button)),)
    
    def show_edit(self):
        self.children=(widgets.HBox((self.label, self.confirm_edit_button, self.escape_edit_button)),widgets.HBox((self.edit_label_value,)))

    def hide_edit(self):
        self.children=(widgets.HBox((self.label, self.edit_button, self.delete_button)),)


def get_head_label(text):
    return widgets.Label(value=text, layout=widgets.Layout(width='250px', height='30px', justify_content='center'))


# this should of course be changed, but i will leave it for now
def get_some_other_label_that_i_dont_know_what_it_is():
    return widgets.Label(value=f"", layout=widgets.Layout(width='250px', height='70px', align_items='stretch'))


class AddLabelBox(widgets.HBox):
    def __init__(self):
        super().__init__()
        self.add_new_label_button = widgets.Button(description="",
                                                   layout=widgets.Layout(width='35px', height='35px'), icon="plus")
        self.label_name_text_box = widgets.Textarea(placeholder='Label name',
                                                    layout=widgets.Layout(width='215px', height='35px'))
        self.children = (self.label_name_text_box, self.add_new_label_button)

class EditLabelBox(widgets.HBox):
    def __init__(self):
        super().__init__()
        self.add_new_label_button = widgets.Button(description="",
                                                   layout=widgets.Layout(width='35px', height='35px'), icon="plus")
        self.label_name_text_box = widgets.Textarea(placeholder='New label name',
                                                    layout=widgets.Layout(width='215px', height='35px'))
        self.children = (self.label_name_text_box, self.add_new_label_button)



def get_labels_info_scrollable():
    return widgets.Output(layout={'overflow_y': 'scroll', 'height': '450px'})



import random

import networkx as nx
import numpy as np

from .subscribe import subscribable


class VisualGraph:
    def __init__(self, graph: nx.Graph, bounds: (int, int)):
        self.graph = graph
        self.bounds = bounds

        self.edge_labels = set()
        self.vertex_labels = set()
        self.edge_edit = True
        self.vertex_edit = True

        self.coordinates = {
            node: [random.randint(0, bounds[0] - 1), random.randint(0, bounds[1] - 1)]
            for node in graph.nodes
        }
        self.selected_node = None
        self.selected_edge = None
        self.dragged_node = None

    @subscribable
    def add_node(self, node, pos: (int, int)):
        self.graph.add_node(node, **dict.fromkeys(self.vertex_labels, ""))
        self.coordinates[node] = pos

    @subscribable
    def add_edge(self, node1, node2):
        self.graph.add_edge(node1, node2, **dict.fromkeys(self.edge_labels, ""))

    @subscribable
    def remove_node(self, node):
        self.graph.remove_node(node)
        del self.coordinates[node]

    @subscribable
    def remove_edge(self, node1, node2):
        self.graph.remove_edge(node1, node2)

    def new_node_label(self, label):
        if label in self.vertex_labels:
            return
        else:
            self.vertex_labels.add(label)
            nx.set_node_attributes(self.graph, "", label)

    def new_edge_label(self, label):
        if label in self.edge_labels:
            return
        else:
            self.edge_labels.add(label)
            nx.set_edge_attributes(self.graph, "", label)
    
    def remove_edge_label(self, label):
        if not label in self.edge_labels:
            return
        else:
            self.edge_labels.remove(label)
            for v1, v2, d in self.graph.edges(data=True):
                del d[label]
            
    def remove_vertex_label(self, label):
        if not label in self.vertex_labels:
            return
        else:
            self.vertex_labels.remove(label)
            for v, d in self.graph.nodes(data=True):
                del d[label]
    
    def edit_edge_label(self, old_label, new_label):
        if (not old_label in self.edge_labels) or new_label in self.edge_labels:
            return
        else:
            self.edge_labels.remove(old_label)
            self.edge_labels.add(new_label)
            for v1,v2,d in self.graph.edges(data=True):
                d[new_label]=d[old_label]
                del d[old_label]

    def edit_node_label(self, old_label, new_label):
        if (not old_label in self.vertex_labels) or new_label in self.vertex_labels:
            return
        else:
            self.vertex_labels.remove(old_label)
            self.vertex_labels.add(new_label)
            for v, d in self.graph.nodes(data=True):
                d[new_label]=d[old_label]
                del d[old_label]

    
    def label_edge(self, edge, label, value):
        if label not in self.edge_labels:
            raise ValueError("Attribute for the label was not set")
        else:
            self.graph.edges[edge][label] = value

    def label_node(self, node, label, value):
        if label not in self.node_labels:
            raise ValueError("Attribute for the label was not set")
        else:
            self.graph.nodes[node][label] = value

    @subscribable
    def move_node(self, node, pos: (int, int)):
        if node not in self.graph.nodes:
            raise ValueError("Node not in graph")
        self.coordinates[node] = pos

    @subscribable
    def drag_start(self, node):
        self.dragged_node = node

    @subscribable
    def drag_end(self):
        self.dragged_node = None

    def get_closest_node(self, pos: (int, int)) -> (any, float):
        closest_node = None
        closest_dist = float("inf")
        for node, node_pos in self.coordinates.items():
            dist = np.linalg.norm(np.array(pos) - node_pos)
            if dist < closest_dist:
                closest_dist = dist
                closest_node = node
        return closest_node, closest_dist

    def get_closest_edge(self, pos: (int, int)) -> (any, float):
        closest_edge = None
        closest_dist = float("inf")
        for u, v in self.graph.edges:
            p1 = np.array(self.coordinates[u])
            p2 = np.array(self.coordinates[v])
            p3 = np.array(pos)
            if np.dot(p3 - p1, p2 - p1) > 0 and np.dot(p3 - p2, p1 - p2) > 0:
                dist = np.linalg.norm(np.cross(p2 - p1, p3 - p1) / np.linalg.norm(p2 - p1))
            else:
                dist = min(np.hypot(*(p3 - p1)), np.hypot(*(p3 - p2)))
            if dist < closest_dist:
                closest_dist = dist
                closest_edge = (u, v)
        return closest_edge, closest_dist



class Subscribable:
    def __init__(self):
        self.subscribers = []

    def subscribe(self, func):
        self.subscribers.append(func)

    def unsubscribe(self, func):
        self.subscribers.remove(func)

    def notify(self, *args, **kwargs):
        for subscriber in self.subscribers:
            subscriber(*args[1:], **kwargs)


def subscribable(func):
    def wrapper(*args, **kwargs):
        wrapper.subscribable.notify(*args, **kwargs)
        result = func(*args, **kwargs)
        return result

    wrapper.subscribable = Subscribable()
    return wrapper


import threading
import time
import graphics
import ipywidgets as widgets
import networkx as nx
from IPython.display import display
from ipycanvas import Canvas
from ipyevents import Event
from .graph_physics import GraphPhysics
from .settings import NODE_CLICK_RADIUS, EDGE_CLICK_RADIUS
from .visual_graph import VisualGraph
from functools import partial
from enum import Enum


class Mode(Enum):
    STRUCTURE = 0
    PROPERTIES = 1


def mex(arr):
    result = 0
    while result in arr:
        result += 1
    return result

def edit(graph: nx.Graph):
    """
    An interactive graph editor function designed for Jupyter.

    This function enables users to manipulate a graph by creating vertices, edges and adding labels.


    Parameters
    ----------
    - graph (networkx.Graph): The graph object to be edited. It should be an instance of the
      NetworkX Graph class or a subclass.

    Functions of buttons in order from left to right
    ------------------------------------------------
    1-2: Select whether you want to edit graph structure (1) or labels (2).
    3. Select if you want for nodes to be clickable.
       Deselecting this should make easier to operate on edges in a large graph.
    4. Select if you want for edges to be clickable.
       Deselecting this should make easier to operate on nodes in a large graph.
    5. Close editing window.

    Mouse functions
    ---------------
    1. Click and drag vertices to move them around the canvas.
    2. To create an edge, click on one vertex and then click on another vertex.
       An edge will be created between the two selected vertices.
    3. To create a vertex, click on empty space of a canvas.
    3. To delete an object, double-click on it.

    Dependencies
    ------------
    - Jupyter notebook web environment.
    - NetworkX library for graph manipulation.

    Notes
    -----
    This function relies on Jupyter ipywidgets, so it should work only in web versions of Jupyter.

    Examples
    --------
    >>> import networkx as nx

    Create a sample graph

    >>> G = nx.Graph()
    >>> G.add_nodes_from([1, 2, 3])
    >>> G.add_edges_from([(1, 2), (2, 3)])

    Call the interactive graph editor

    >>> edit(G)
    """
    # logical properties of the graph
    ################################
    visual_graph = VisualGraph(graph, (800, 500))
    CLOSE = False
    mode = Mode.STRUCTURE
    is_drag = False
    start_mouse_position = (0, 0)
    actions_to_perform = []
    EPS = 10
    #################################
    #################################

    canvas = Canvas(width=800, height=500)

    # main menu
    #################################
    mode_box = graphics.Menu()

    def close(button):
        nonlocal CLOSE, main_box
        CLOSE = True
        main_box.children = ()

    mode_box.close_button.on_click(close)

    def click_struct(new_mode, button_widget):
        nonlocal mode, mode_box
        if not button_widget.active:
            mode = new_mode
            mode_box.struct_button.toggle()
            mode_box.prop_button.toggle()
        update_labels(labels_info, visual_graph)

    mode_box.struct_button.on_click(partial(click_struct, Mode.STRUCTURE))
    mode_box.prop_button.on_click(partial(click_struct, Mode.PROPERTIES))

    def click_verts_select(button_widget):
        nonlocal visual_graph
        button_widget.toggle()
        if not button_widget.active:
            visual_graph.selected_node = None
            update_labels(labels_info, visual_graph)

    mode_box.vert_button.on_click(click_verts_select)

    def click_edge_select(button_widget):
        nonlocal visual_graph
        button_widget.toggle()
        if not button_widget.active:
            visual_graph.selected_edge = None
            update_labels(labels_info, visual_graph)

    mode_box.edge_button.on_click(click_edge_select)
    #######################

    # labels
    #######################

    labels_info = widgets.VBox()
    add_label_box = graphics.AddLabelBox()

    labels_info_scrollable = graphics.get_labels_info_scrollable()
    with labels_info_scrollable:
        display(labels_info)

    def add_label(button_widget, labels_info: widgets.VBox, visual_graph: VisualGraph, label_name: widgets.Textarea):
        new_label_name = str(label_name.value)
        if visual_graph.selected_node is not None:
            if new_label_name in visual_graph.vertex_labels:
                return
            else:
                visual_graph.new_node_label(new_label_name)

            update_labels(labels_info=labels_info, visual_graph=visual_graph)

        elif visual_graph.selected_edge is not None:
            if new_label_name in visual_graph.edge_labels:
                return
            else:
                visual_graph.new_edge_label(new_label_name)

            update_labels(labels_info=labels_info, visual_graph=visual_graph)

    on_click = partial(add_label, labels_info=labels_info, visual_graph=visual_graph,
                       label_name=add_label_box.label_name_text_box)
    add_label_box.add_new_label_button.on_click(on_click)

    def update_labels(labels_info: widgets.VBox, visual_graph: VisualGraph):
        nonlocal mode, add_label_box
        add_label_box.label_name_text_box.value=''
        if mode is Mode.PROPERTIES:
            if visual_graph.selected_node is not None:
                head_text = f"Node {repr(visual_graph.selected_node)}"
                labels_info.children = (graphics.get_head_label(head_text),)

                for i in visual_graph.graph.nodes[visual_graph.selected_node].keys():
                    value = str(visual_graph.graph.nodes[visual_graph.selected_node][i])
                    new_label = graphics.LabelBox(str(i), value)

                    def modify_label(change, visual_graph: VisualGraph):
                        visual_graph.graph.nodes[visual_graph.selected_node][i] = change["new"]

                    on_change = partial(modify_label, visual_graph=visual_graph)
                    new_label.label_value.observe(on_change, names="value")
                    labels_info.children += (new_label,)

                labels_info.children += (widgets.VBox([add_label_box]),)

            elif visual_graph.selected_edge is not None:
                head_text = f"Edge {repr(visual_graph.selected_edge)}"
                labels_info.children = (graphics.get_head_label(head_text),)

                for i in visual_graph.graph.edges[visual_graph.selected_edge].keys():
                    value = str(visual_graph.graph.edges[visual_graph.selected_edge][i])
                    new_label = graphics.LabelBox(str(i), value)
                    def modify_label(change, visual_graph: VisualGraph):
                        visual_graph.graph.edges[visual_graph.selected_edge][i] = change["new"]

                    on_change = partial(modify_label, visual_graph=visual_graph)
                    new_label.label_value.observe(on_change, names="value")
                    labels_info.children += (new_label,)
                labels_info.children += (widgets.VBox([add_label_box]),)

            else:
                labels_info.children = (graphics.get_head_label(f"Node labels: "),)
                def edit(button, name_label):
                    name_label.show_edit()

                def escape_edit(button, name_label):
                    name_label.hide_edit()

                def remove_vertex_label(button, visual_graph, label, labels_info):
                    visual_graph.remove_vertex_label(label)
                    update_labels(labels_info=labels_info, visual_graph=visual_graph)

                def edit_vertex_label(button, visual_graph, name_label, old_name, labels_info):
                    visual_graph.edit_node_label(old_label=old_name, new_label=str(name_label.edit_label_value.value))
                    update_labels(labels_info=labels_info, visual_graph=visual_graph)

                for name in visual_graph.vertex_labels:
                    name_label= graphics.LabelListBox(name)
                    name_label.delete_button.on_click(partial(remove_vertex_label, label=name,visual_graph=visual_graph, labels_info=labels_info))
                    name_label.edit_button.on_click(partial(edit, name_label=name_label))
                    name_label.escape_edit_button.on_click(partial(escape_edit,name_label=name_label))
                    name_label.confirm_edit_button.on_click(partial(edit_vertex_label, visual_graph=visual_graph, name_label=name_label, old_name=name, labels_info=labels_info))
                    labels_info.children += (name_label,)

                def remove_edge_label(button,visual_graph, label, labels_info):
                    visual_graph.remove_edge_label(label)
                    update_labels(labels_info=labels_info, visual_graph=visual_graph)

                def edit_edge_label(button, visual_graph, name_label, old_name, labels_info):
                    visual_graph.edit_edge_label(old_label=old_name, new_label=str(name_label.edit_label_value.value))
                    update_labels(labels_info=labels_info, visual_graph=visual_graph)

                labels_info.children += (graphics.get_head_label(f"Edge labels: "),)
                for name in visual_graph.edge_labels:
                    name_label= graphics.LabelListBox(name)
                    name_label.delete_button.on_click(partial(remove_edge_label, label=name, visual_graph=visual_graph, labels_info=labels_info))
                    name_label.edit_button.on_click(partial(edit, name_label=name_label))
                    name_label.escape_edit_button.on_click(partial(escape_edit, name_label=name_label))
                    name_label.confirm_edit_button.on_click(partial(edit_edge_label, visual_graph=visual_graph, name_label=name_label, old_name=name, labels_info=labels_info))
                    labels_info.children += (name_label,)
        else:
            labels_info.children = (graphics.get_some_other_label_that_i_dont_know_what_it_is(),)

    ##############################

    # canvas actions
    ##############################

    def node_click(node):
        visual_graph.selected_edge = None
        if visual_graph.selected_node is None or visual_graph.selected_node != node:
            visual_graph.selected_node = node
        else:
            visual_graph.selected_node = None

    def edge_click(edge):
        visual_graph.selected_node = None
        if visual_graph.selected_edge is None or visual_graph.selected_edge != edge:
            visual_graph.selected_edge = edge
        else:
            visual_graph.selected_edge = None

    def handle_mousedown(event):
        nonlocal mode
        nonlocal start_mouse_position
        start_mouse_position = (event['relativeX'], event['relativeY'])
        if mode is Mode.PROPERTIES:
            if mode_box.vert_button.active:
                clicked_node, dist = visual_graph.get_closest_node((event['relativeX'], event['relativeY']))
                if dist < NODE_CLICK_RADIUS:
                    node_click(clicked_node)
                    update_labels(labels_info, visual_graph)
                    return

            if mode_box.edge_button.active:
                clicked_edge, dist = visual_graph.get_closest_edge((event['relativeX'], event['relativeY']))
                if dist < EDGE_CLICK_RADIUS:
                    visual_graph.selected_node = None
                    # we will select the edge also when dragging, this behaviour can be changed
                    edge_click(clicked_edge)
                    update_labels(labels_info, visual_graph)
                else:
                    visual_graph.selected_edge=None
                    visual_graph.selected_node=None
                    update_labels(labels_info, visual_graph)

        else:
            if mode_box.vert_button.active:
                clicked_node, dist = visual_graph.get_closest_node((event['relativeX'], event['relativeY']))
                if dist < NODE_CLICK_RADIUS:
                    visual_graph.drag_start(clicked_node)
                    visual_graph.selected_edge = None
                    return

    def handle_mousemove(event):
        nonlocal mode, is_drag, EPS
        distance = abs(start_mouse_position[0] - event['relativeX']) + abs(start_mouse_position[1] - event['relativeY'])
        if mode is Mode.STRUCTURE:
            nonlocal is_drag
            if visual_graph.dragged_node is not None and distance > EPS:
                is_drag = True
                pos = (event['relativeX'], event['relativeY'])
                visual_graph.move_node(visual_graph.dragged_node, pos)

    def handle_mouseup(event):
        nonlocal mode
        # handling in PROPERTIES mode is handled with clicking, as we cannot drag in that mode
        if mode is Mode.STRUCTURE:
            nonlocal is_drag
            visual_graph.drag_end()
            if is_drag:
                is_drag = False
                return

            pos = (event['relativeX'], event['relativeY'])
            # selecting edges is handled with clicking, as we canot drag them
            # perhaps we should add the functionality of dragging edges as well and change that
            if mode_box.vert_button.active:
                node, dist = visual_graph.get_closest_node(pos)
                if dist < NODE_CLICK_RADIUS:
                    node = visual_graph.get_closest_node(pos)[0]

                    if visual_graph.selected_node is None:
                        visual_graph.selected_node = node
                        update_labels(labels_info, visual_graph)

                    elif visual_graph.selected_node == node:
                        visual_graph.selected_node = None
                        update_labels(labels_info, visual_graph)

                    elif not visual_graph.graph.has_edge(visual_graph.selected_node, node):
                        visual_graph.add_edge(visual_graph.selected_node, node)
                    else:
                        visual_graph.selected_node = node
                        update_labels(labels_info, visual_graph)
                    return

            if mode_box.edge_button.active:
                clicked_edge, dist = visual_graph.get_closest_edge((event['relativeX'], event['relativeY']))
                if dist < EDGE_CLICK_RADIUS:
                    # we will select the edge also when dragging, this behaviour can be changed
                    # for now the only thing that selecting an edge will do will be showing its properties
                    edge_click(clicked_edge)
                    update_labels(labels_info, visual_graph)
                    return

            # if we didn't click vertex nor edge
            if visual_graph.selected_node is None and visual_graph.selected_edge is None:
                new_node = mex(visual_graph.graph.nodes)
                visual_graph.add_node(new_node, pos)
            else:
                visual_graph.selected_node = None
                visual_graph.selected_edge = None
                update_labels(labels_info, visual_graph)

    def handle_doubleclick(event):
        nonlocal mode
        if mode is Mode.STRUCTURE:
            pos = (event['relativeX'], event['relativeY'])
            if mode_box.vert_button.active:
                clicked_node, dist = visual_graph.get_closest_node(pos)
                if dist < NODE_CLICK_RADIUS:
                    visual_graph.remove_node(clicked_node)
                    visual_graph.selected_node = None
                    return

            if mode_box.edge_button.active:
                clicked_edge, dist = visual_graph.get_closest_edge(pos)
                if dist < EDGE_CLICK_RADIUS:
                    visual_graph.selected_edge = None
                    visual_graph.remove_edge(clicked_edge[0], clicked_edge[1])
                    debug_text.value = str(clicked_edge)

    def perform_in_future(action):
        def event_consumer(*args, **kwargs):
            actions_to_perform.append((action, args, kwargs))

        return event_consumer

    Event(source=canvas, watched_events=['mousedown']).on_dom_event(perform_in_future(handle_mousedown))
    Event(source=canvas, watched_events=['mousemove'], wait=1000 // 60).on_dom_event(
        perform_in_future(handle_mousemove))
    Event(source=canvas, watched_events=['mouseup']).on_dom_event(perform_in_future(handle_mouseup))
    Event(source=canvas, watched_events=['dblclick']).on_dom_event(perform_in_future(handle_doubleclick))

    ##################################
    ##################################

    # main structure and main loop
    #############################

    main_box = widgets.HBox()
    debug_text = widgets.Textarea()

    main_box.children = ([widgets.VBox((mode_box, labels_info_scrollable)), canvas])
    display(main_box)

    #display(output)#for debugging
    #display(debug_text)
    update_labels(labels_info, visual_graph)
    graph_physics = GraphPhysics(visual_graph)

    def main_loop(visual_graph, physics_button):
        nonlocal CLOSE
        try:
            while not CLOSE:
                graph_physics.update_physics(1 / 60, physics_button.value)
                graph_physics.normalize_positions()
                graphics.draw_graph(canvas, visual_graph)
                time.sleep(1 / 60)
                for (action, args, kwargs) in actions_to_perform:
                    action(*args, **kwargs)
                actions_to_perform.clear()
        except Exception as e:
            debug_text.value = repr(e)

    thread = threading.Thread(target=main_loop, args=(visual_graph, mode_box.physics_button))
    thread.start()

    ################################
    ################################
